"""
    Utilities for OS level filesystem checks.
"""
import re
import sys
import subprocess

import os.path as osp

# TODO: refactor
# idea 1: into check_path and check_files
# idea 2:
def os_filesystem_check(directory, files_list, filetype_pattern):
    """
        Checks if file types match pattern.
        Input:
            directory = str : path to directory, eg. "/proc/net"
            files_list = list : list of str, eg. "["f1", "f2"]"
            filetype_pattern = str: eg. " empty$"
    """
    passed = True

    # TODO: import struct?
    class CHECK_RESULTS:
        """
            ANSI escape sequences for colored output.
        """
        OK = '\033[92m OK \033[0m'
        FAIL = '\033[91m FAIL \033[0m'


    cdirpath = directory.rstrip('/') + '/'

    flist_format = []
    for file_name in files_list:
        flist_format.append(cdirpath + file_name.strip('/'))

    # directory check, don't check files if this fails
    sys.stdout.write('Checking path: ' + cdirpath.ljust(24) + ' ... ')
    if not osp.exists(cdirpath.rstrip('/')):
        # unlike isdir(), exists() fails with trailing '/'
        sys.stdout.write(CHECK_RESULTS.FAIL + '\n')
        sys.stderr.write('Path ' + cdirpath + ' does not exist or not accessible. Check OS.\n')
        return not passed
    elif not osp.isdir(cdirpath):
        sys.stdout.write(CHECK_RESULTS.FAIL + '\n')
        sys.stderr.write('Path ' + cdirpath + ' is not a directory. \n')
        return not passed
    else:
        sys.stdout.write(CHECK_RESULTS.OK + '\n')

    # iterate files_list, continue on verification fail with continue kw
    # files_list will all be checked under cdirpath
    for file_n in flist_format:
        cfile_path = file_n
        r_status = CHECK_RESULTS.OK

        # file in question
        sys.stdout.write('Checking file: ' + cfile_path.ljust(24) + ' ... ')

        # file type matches pattern, see Documentation/Popen/file.c
        p_proc = subprocess.Popen(['stat', cfile_path], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        subp_out, subp_err = p_proc.communicate()

        subp_out = subp_out.decode('utf-8')
        subp_err = subp_err.decode('utf-8')

        if p_proc.returncode != 0:
            passed = False
            r_status = CHECK_RESULTS.FAIL
            sys.stdout.write(r_status + '\n')
            sys.stderr.write('There were errors verifying type of ' + cfile_path + ' ... \n' + subp_err + '\n')
            continue

        pattern_filetype = re.compile(r'' + str(filetype_pattern))

        #DEBUG
        # print(subp_out)
        # print(type(subp_out))
        #/dEBUG


        regex_result = pattern_filetype.search(str(subp_out))
        if not regex_result:
            passed = False
            r_status = CHECK_RESULTS.FAIL
            sys.stdout.write(r_status + '\n')
            sys.stderr.write('File type of ' + cfile_path + ' does not correspond. \n')
            continue

        # depth level 0, no if statement passed
        sys.stdout.write(r_status + '\n')

    return passed



"""
    Network related operations.
    TODO: master class with static methods and os checks for connectivity
"""

import socket
import whois
import time


def resolve_hostname(dest_ip):
    """
        Try to query hostname.
    """

    #limit_flood = 0.116
    #TODO: pass from __main__
    #time.sleep(limit_flood)

    try:
        return str(socket.gethostbyaddr(dest_ip)[0])
    except socket.herror as he:
        return str("")
    except socket.gaierror as ge:
        raise ge

def resolve_location(hostname, verbose=True):
    """
        Try to assign a location to a hostname.
    """
    whois_response = None
    geoip_response = None # TODO: geoip
    tld_string = None


    whois_countries = dict()

    rloc = None

    for i in range(3):
        if verbose:
            print("Attempting to resolve location: "  + hostname)

        try:
            whois_response = whois.whois(hostname)
        except Exception as ex:
            print(ex)
            continue

        tdata = whois_response
        try:
            tdata = tdata.text
        except AttributeError as ae:
            continue


        cdata = []
        for i in tdata.split('\n'):
            if 'Registrant Country:' in i:
                ccode = i.split(' ')[2].rstrip('\r').rstrip('\n').rstrip()
                ccode = ccode.lower()
                cdata.append(ccode)
                whois_countries[ccode] = 1 if ccode not in whois_countries else whois_countries[ccode] + 1

        break


    tld_string = hostname.split('.')[-1].rstrip().lstrip('.')

    if len(whois_countries.keys()) != 0:
        rloc = list(whois_countries.keys())[0]
    #elif geoip_response != None:
    #    rloc = geoip_response
    elif tld_string != None:
        return tld_string

    return rloc



"""
    File parser and syntax analysis class, for /proc/net/tcp.
"""

import re
import os
import sys

import tempfile

import pygal

from network_controller import resolve_hostname, resolve_location

from tcp_file_entry import EntryTCP
from tcp_structs_exceptions import  EntryTCP_Error, \
                                    EntryTCP_FormatError, \
                                    FileTCP_InitError


import socket # TODO: bad exception handling with socket.gaierror

class FileTCP(object):
    """
        File data parser.
    """
    def __init__(self, path):
        self.re_path_mchr = re.compile(r'[a-zA-Z0-9\_\/]') # TODO: practice?
        self.path = path
        self.data = None
        self.entries = None
        self.removed_lines = 0

        self.entry_locations = dict()
        self.entry_hostnames = dict()
        self.tempfile_name = None
        self.tempfile_handler = None
        self.last_write_filename = None

        if not isinstance(path, str) or path == "":
            raise FileTCP_InitError("Not a valid path string: ", "{0}".format(path))

        for s_chr in path:
            if not self.re_path_mchr.match(s_chr):
                raise FileTCP_InitError('Path contains invalid characters.', s_chr) # TODO: exception style

    def __del__(self):
        if self.tempfile_name is not None:
            os.remove(self.tempfile_name)

    def parse_entries(self):
        """
            If self.data is populated with the /net/tcp text file's read data,
            it is analyzed by parse_entries for syntax checking, and passing
            lines will be converted into EntryTCP objects and stored in object
            self.entries.
        """
        # TODO: InitError as superclass of FileTCP_NoDataError, msg len
        # TODO: exception for empty file? tests for empty file? test for connectivity?
        if self.data is None:
            raise FileTCP_InitError("Fatal error, no data to parse.", "Tried with path: {0}".format(self.path))

        self.entries = self.data.split("\n")

        # delete non-entry lines
        while True:
            done = True
            for index in range(len(self.entries)):
                if not isinstance(self.entries[index], EntryTCP):
                    if not isinstance(self.entries[index], str) or self.entries[index] == "":
                        # these lines do not pass to EntryTCP constructor
                        self.removed_lines += 1
                        del self.entries[index]
                        done = False
                        break


                    try:
                        self.entries[index] = EntryTCP(self.entries[index])
                    except EntryTCP_FormatError:
                        # lines with incorrect format are removed
                        self.removed_lines += 1
                        del self.entries[index]
                        done = False
                        break

            if done:
                break

    def read_tcp_struct(self, data=None):
        """
            Initializes self.data property with either function argument or from
            file data pointed to by self.path.
        """
        if data is None:
            try:
                with open(self.path, 'r') as tcp_fd:
                    self.data = tcp_fd.read()
            except IOError as io_e:
                # Errno 2 not exist
                sys.stderr.write(str(io_e))
                sys.stderr.write("Error, file not accessible.\n")
        else:
            self.data = data

        if self.data is None:
            sys.stderr.write("Nothing was read...")
        else:
            self.parse_entries()

    def get_entries(self):
        """
            Getter for public field.
        """
        return self.entries


    def draw_map_v3(self, mode=None, continuous = None):
        for entry in self.entries:
            if entry.resolved_location != None:
                self.entry_locations[entry.dest_ip] = entry.resolved_location
            else:
                if entry.resolved_location == None:
                    entry.resolve_country()
                if entry.resolved_location != None:
                    self.entry_locations[entry.dest_ip] = entry.resolved_location


        worldmap_chart = pygal.maps.world.World()
        worldmap_chart.title = 'Some countries'

        if mode == None or mode == 'flag':
            for item in self.entry_locations.keys():
                val = self.entry_locations[item]
                worldmap_chart.add(item, val)
        elif mode == 'heatmap':
            worldmap_chart.add('Heatmap', self.entry_locations.values())


        if self.tempfile_name == None:
            new_file, filename = tempfile.mkstemp()
            self.tempfile_name = filename
            self.last_write_name = self.tempfile_name

            # todo: remove handler
            self.tempfile_handler = new_file
            os.close(new_file)
            worldmap_chart.render_to_png(self.tempfile_name)
            if not continuous:
                self.tempfile_name = None
        else:
            worldmap_chart.render_to_png(self.tempfile_name)
            self.last_write_name = self.tempfile_name
            if not continuous:
                self.tempfile_name = None


    def draw_map_v2(self, mode=None, continuous=None):
        """
            Generate map from stored entries property.
        """
        for entry in self.entries:

            if str(entry.dest_ip) not in self.entry_hostnames:
                rhn = ""
                try:
                    rhn = resolve_hostname(str(entry.dest_ip))
                except socket.gaierror as ge:
                    rhn = "UNKNOWN_HOSTNAME"
                if rhn != "":
                    self.entry_hostnames[str(entry.dest_ip)] = rhn

            if str(str(entry.dest_ip)) not in self.entry_locations:
                if str(str(entry.dest_ip)) in self.entry_hostnames:
                    self.entry_locations[str(entry.dest_ip)] = resolve_location(self.entry_hostnames[str(entry.dest_ip)])

        worldmap_chart = pygal.maps.world.World()
        worldmap_chart.title = 'Some countries'

        if mode == None or mode == 'flag':
            for item in self.entry_locations.keys():
                val = self.entry_locations[item]
                worldmap_chart.add(item, val)
        elif mode == 'heatmap':
            worldmap_chart.add('Heatmap', self.entry_locations.values())


        if self.tempfile_name == None:
            new_file, filename = tempfile.mkstemp()
            self.tempfile_name = filename
            self.last_write_name = self.tempfile_name

            self.tempfile_handler = new_file
            os.close(new_file)
            worldmap_chart.render_to_png(self.tempfile_name)
            if not continuous:
                self.tempfile_name = None
        else:
            worldmap_chart.render_to_png(self.tempfile_name)
            self.last_write_name = self.tempfile_name
            if not continuous:
                self.tempfile_name = None


    def print_entries(self, resolve=False):
        """
            Prints EntryTCP objects to stdout, one item per line.
        """
        for item in self.entries:
            if not resolve:
                sys.stdout.write(str(item))
            else:
                resolved_host = ""
                try:
                    resolved_host = resolve_hostname(str(item.dest_ip))
                except socket.gaierror as ge:
                    resolved_host = "UNKNOWN_HOSTNAME"
                sys.stdout.write(str(item).rstrip() + " \t" + str(resolved_host) + " "  + " " + '\n')



"""
    Defines connection state structs as specified in the Linux kernel.
    https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/tree/include/net/tcp_states.h
    Accessed from /proc/net/tcp.
"""

class C_STATE:
    established     = 0x1
    syn_sent        = 0x2
    syn_recv        = 0x3
    fin_wait1       = 0x4
    fin_wait2       = 0x5
    time_wait       = 0x6
    close           = 0x7
    close_wait      = 0x8
    last_ack        = 0x9
    listen          = 0xA
    tcp_closing     = 0xB

    strings = {
        0x1: "established",
        0x2: "syn_sent",
        0x3: "syn_recv",
        0x4: "fin_wait1",
        0x5: "fin_wait2",
        0x6: "time_wait",
        0x7: "close",
        0x8: "close_wait",
        0x9: "last_ack",
        0xA: "listen",
        0xB: "tcp_closing"
    }

    lower_bound = 0x1
    upper_bound = 0xB

    @staticmethod
    def check_bounds(value):
        """
            Verifies if value inside argument is within class bounds.
            Returns the same value.
            Exceptions:
                ValueError
        """
        sv = C_STATE.lower_bound
        ev = C_STATE.upper_bound
        if value < sv or value > ev:
            raise ValueError("Value provided was out of bounds.")

        return value

    @staticmethod
    def string_from_hex(value):
        """
            Returns connection type name, defined the C_STATE class or by
            net/tcp_states.h in the Linux kernel.
        """
        C_STATE.check_bounds(value)
        return C_STATE.strings[value] if value in C_STATE.strings else None

    @staticmethod
    def hex_from_string(string):
        """
            Returns value correlated to the connection state string.
            Defined by C_STATE class, or net/tcp_states.h
        """
        if type(string) is not str:
            raise TypeError("Function requires str as argument.")

        for value in range(C_STATE.lower_bound, C_STATE.upper_bound + 1):
            if C_STATE.strings[value] == string:
                return value

        return None



"""
    Entry point.
"""

import contextlib
with contextlib.redirect_stdout(None):
    import pygame

import signal
import argparse
import time
import os
import sys

sys.path.append(os.path.realpath(os.path.dirname(__file__)+"/.."))
sys.path.append(os.path.realpath(os.path.dirname(__file__)+"/../distnet"))

from network_controller import resolve_hostname
from tcp_file_handler import FileTCP
from os_check import os_filesystem_check


import pkg_resources

versionFile = pkg_resources.resource_filename(__name__, 'resources/VERSION')

__version__ = None

with open(versionFile, 'r') as fd:
    __version__ = fd.read()
    fd.close()


def signal_handler(sig, frame):
        print('You pressed Ctrl+C!')
        running = False
        sys.exit(0)


#TODO: refactor into game controller class
def world(x, y, gameDisplay_obj, img_location):
    try:
        gameDisplay_obj.blit(pygame.image.load(img_location), (x, y))
    except pygame.error as pe:
        print('Image draw error.')

def main():

    signal.signal(signal.SIGINT, signal_handler)

    parser = argparse.ArgumentParser()

    parser.add_argument("-t", "--test", action="store_true", help="test os filesystem")
    parser.add_argument("-v", "--verbose", action="store_true", help="verbose output")
    parser.add_argument("--visual", action="store_true", help="draw on world map (use with -c)")
    parser.add_argument("-c", "--continuous", action="store_true", help="continuously update output")
    parser.add_argument("--version", action="store_true", help="prints version of this package")
    parser.add_argument("--output", action="append", help="file to write to", type=argparse.FileType('w'))
    parser.add_argument("-r", "--resolve", action="store_true", help="resolve ip addresses to hostnames")


    args = parser.parse_args()

    if args.version == True:
        print("Version is: " + str(__version__))
        exit(0)

    if args.test == True:
        if os_filesystem_check(directory='/proc/net///', files_list=[
            '/tcp',
            '//tcp6',
            'udp',
            'udp6'
        ], filetype_pattern=': empty$'):
            with open('/proc/net/tcp', 'r') as ftcp:
                print(ftcp.read())
                print("End file.\n\n")
            with open('/proc/net/tcp6', 'r') as ftcp:
                print(ftcp.read())
                print("End file.\n\n")
        exit(0)

    # 2> /dev/null
    # redir_stderr = open(os.devnull, 'w')
    # sys.stderr = redir_stderr

    time_sleep = 1 # TODO: default-valued int argument argparse

    if args.output != None:
        fd_args = args.output[0]
        sys.stdout = fd_args

    if args.continuous == True:
        history_ips = dict()
        cdict = dict()

        gameDisplay = None
        if args.visual == True:
            pygame.init()
            pygame.mixer.quit()
            gameDisplay = pygame.display.set_mode((800, 600))
            pygame.display.set_caption('Visualizer')

        if args.visual == True:
            clock = pygame.time.Clock()

        ftcp = FileTCP('/proc/net/tcp')
        running = True
        while running:

            ftcp.read_tcp_struct()

            if args.verbose == True:
                print(ftcp.data)

            if args.output == None:

                en = ftcp.get_entries()

                for entry in en:
                    # todo: change if for max_tries
                    if entry.dest_ip not in history_ips.keys():
                        history_ips[entry.dest_ip] = True
                        ns_formatted = ""

                        if args.resolve == True:
                            resolved_hostname = ""
                            try:
                                if entry.resolved_hostname == None:
                                    entry.resolve_country()
                                resolved_hostname = entry.resolved_hostname
                                #resolved_hostname = resolve_hostname(entry.dest_ip)
                            except socket.gaierror as ge:
                                resolved_hostname = "UNKNOWN_HOSTNAME"
                            ns_formatted += " " + resolved_hostname + " "  + " "

                        sys.stdout.write(entry.dest_ip + ns_formatted + '\n')
            else:
                en = ftcp.get_entries()
                for entry in en:
                    if entry.dest_ip not in history_ips.keys():
                        history_ips[entry.dest_ip] = True
                        ns_formatted = ""

                        if args.resolve == True:
                            resolved_hostname = ""
                            try:
                                resolved_hostname = resolve_hostname(entry.dest_ip)
                            except socket.gaierror as ge:
                                resolved_hostname = "UNKNOWN_HOSTNAME"
                            ns_formatted += " " + resolved_hostname + " "  + " "

                        sys.stdout.write(entry.dest_ip + ns_formatted + '\n')

            if args.visual == True:
                ftcp.draw_map_v3(continuous=args.continuous)

                world(0, 0, gameDisplay, ftcp.tempfile_name)
                pygame.display.update()
                clock.tick(24)
                for event in pygame.event.get():
                    if event.type == pygame.QUIT:
                        running = False


            #time.sleep(1)
            sys.stdout.flush()
    else:
        ftcp = FileTCP('/proc/net/tcp')
        ftcp.read_tcp_struct()

        if args.verbose == True:
            print(ftcp.data)

        ftcp.print_entries(resolve=args.resolve)

        if args.visual == True:
            ftcp.draw_map_v2(mode='heatmap', continuous = args.continuous)
            print("Wrote image to: " + ftcp.last_write_name)


    #redir_std.close()
    sys.stdout.write("Done.\n")

    if args.output != None:
        args.output.close()

    exit(0)

if __name__ == '__main__':
    main()



"""
    Whole-project exceptions.
"""
import sys


#TODO: comments and specs
#TODO: restructure exceptions and inheritance

class EntryTCP_Error(Exception):
    def __init__(self, message):
        super(EntryTCP_Error, self).__init__(message)

class EntryTCP_FormatError(EntryTCP_Error):
    def __init__(self, message, case):
        super(EntryTCP_FormatError, self).__init__(message)
        self.message = message
        self.case = case

    def __str__(self):
        sys.stderr.write(self.case)

class EntryTCP_InitError(EntryTCP_Error):
    def __init__(self, message, case):
        super(EntryTCP_InitError, self).__init__(message)
        self.string = str(case)
        sys.stderr.write(self.string)

class HexadecimalStringFormatError(Exception):
    def __init__(self, message, case):
        super(HexadecimalStringFormatError, self).__init__(message)
        self.string = case
        sys.stderr.write(self.string)

class HexadecimalIpFormatError(Exception):
    def __init__(self, message, case):
        super(HexadecimalIpFormatError, self).__init__(message)
        self.string = case
        sys.stderr.write(self.string)

class HexadecimalPortFormatError(Exception):
    def __init__(self, message, case):
        super(HexadecimalPortFormatError, self).__init__(message)
        self.string = case
        sys.stderr.write(self.string)

class FileTCP_Error(Exception):
    def __init__(self, message):
        super(FileTCP_Error, self).__init__(message)

class FileTCP_InitError(FileTCP_Error):
    def __init__(self, message, case):
        super(FileTCP_InitError, self).__init__(message)
        self.string = case
        sys.stderr.write(self.string)

class HostnameNotResolvedError(FileTCP_Error):
    def __init__(self, message, case):
        super(FileTCP_Error, self).__init__(message)
        self.string = case
        sys.stderr.write(self.string)



from setuptools import setup, find_packages
from pkg_resources import resource_filename
# import pkg_resources


__version__ = '0.1.602'

versionFile = resource_filename(__name__, 'distnet/resources/VERSION')

with open(versionFile, 'w') as fd:
    fd.write(__version__)
    fd.close()

setup(
    name='distnet',
    version=__version__,
    author='p0licat',
    author_email='tbodica@gmail.com',
    install_requires=[
        'pygal_maps_world',
        'pygame',
        'python-whois', #TODO: check for whois package, conflicting?
        'cairosvg',
    ],
    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    url='https://github.com/p0licat/distnet',
    license='LICENSE.txt',
    description='TCP connections visualizer.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
        "Operating System :: POSIX :: Linux",
    ],
)



"""
    Manipulate lines in /proc/net/tcp file.
"""


import socket
import network_controller


from hex_manip import ip_from_hex, port_from_hex, int_from_string
from tcp_structs_exceptions import  EntryTCP_FormatError, \
                                    EntryTCP_InitError, \
                                    HostnameNotResolvedError
from tcp_structs import C_STATE


class EntryTCP(object):
    """
        Validated line in TCP file. Supports conversions from hexadecimal.
    """
    def __init__(self, entry_line):

        # entries have fields:
        #   * entry.string      ->  initializer string of EntryTCP constructor
        #   * entry.local_ip    ->  local IP string : re.compile(r'[\.0-9]')
        #   * entry.local_port  ->  local port string : re.compile(r'[0-9]')
        #   * entry.dest_ip     ->  dest IP string : re.compile(r'[\.0-9]')
        #   * entry.dest_port   ->  dest port string : re.compile(r'[0-9]')
        #   * entry.state       ->  0x1 -> 0xB matching C_STATE
        #   * entry.resolved_hostname -> hostname from ns query
        #   * entry.resolved_location -> decision tree result (geoip, whois)


        if not type(entry_line) is str:
            raise EntryTCP_InitError('Entry line must be string. {0}'.format(entry_line), entry_line)

        if entry_line == "":
            raise EntryTCP_FormatError('Entry line must not be null string.', entry_line)

        self.string = entry_line
        if self.string[0] == '\t' or self.string[-1] == '\t' or \
            self.string[0] == ' ' or self.string[-1] == ' ':
            self.string = entry_line.strip('\t ')

        self.local_ip = None
        self.dest_ip = None

        self.local_port = None
        self.dest_port = None

        self.state = None
        self.resolved_hostname = None
        self.resolved_location = None

        self.max_resolve_tries = 2

        issplit = self.string.split(' ')
        ssplit = []

        for item in issplit:
            if item != '':
                ssplit.append(item)

        if len(self.string) < 120:
            raise EntryTCP_FormatError('Line length not recognized as /proc/net format.', "Length was: {0}\n".format(len(self.string)))

        if len(ssplit) < 17:
            raise EntryTCP_FormatError("Columns were less than expected.", 'Number of columns: {0}, with expected: ??\n'.format(len(ssplit)))

        if ssplit[0] == 'sl':
            raise EntryTCP_FormatError("Header line.", "Found header line identifier: {0}\n".format(ssplit[0]))

        #entry_number = ssplit[0]

        """
            See /Documentation for file structure.
            NUMBER  local_address rem_address   st tx_queue rx_queue tr tm->when retrnsmt   uid  timeout inode
        """

        local_addr_field = ssplit[1]
        dest_addr_field = ssplit[2]
        conn_state = ssplit[3]

        s_addr_field = local_addr_field.split(':')
        ip_hex = s_addr_field[0]
        str_port = s_addr_field[1]

        ip_dec = ip_from_hex(ip_hex)
        port = port_from_hex(str_port)

        self.local_ip = ip_dec
        self.local_port = port

        s_addr_field = dest_addr_field.split(':')
        ip_hex = s_addr_field[0]
        str_port = s_addr_field[1]

        ip_dec = ip_from_hex(ip_hex)
        port = port_from_hex(str_port)

        self.dest_ip = ip_dec
        self.dest_port = port

        self.state = conn_state

        #TODO: exceptoins some time
        self.resolved_hostname = self.resolve_hostname()

        try:
            self.resolved_location = self.resolve_country()
        except HostnameNotResolvedError:
            print("Not resolved hostname, so can't check where")


    def resolve_hostname(self):
        resolved_str = None
        if self.max_resolve_tries < 0:
            return
        try:
            resolved_str = network_controller.resolve_hostname(self.dest_ip)
        except socket.gaierror as ge:
            resolved_str = None
        self.resolved_hostname = resolved_str
        self.max_resolve_tries -= 1

    def resolve_country(self):
        if self.resolved_hostname != None:
            self.resolved_location = network_controller.resolve_location(self.resolved_hostname)
        else:
            self.resolve_hostname()
            if self.resolved_hostname == None:
                raise HostnameNotResolvedError('Hostname was not resolved by dns... please retry', str(self.dest_ip))

    def __str__(self):
        rstr = ""
        rstr += "state: " + str(C_STATE.string_from_hex(int_from_string(self.state))) + " \t\t"
        rstr += "from: " + str(self.local_ip) + ":" + str(self.local_port) + " \t\t"
        rstr += "dest: " + str(self.dest_ip) + ":" + str(self.dest_port) + " \t\t"
        rstr += '\n'
        return rstr



"""
    Module contains classes and methods for hexadecimal conversions between str,
    hexadecimal, and formats found in /proc/net/ files in Linux.
"""


from tcp_structs_exceptions import HexadecimalIpFormatError, \
    HexadecimalPortFormatError, HexadecimalStringFormatError


def hex_dict():
    """
        Returns dictionary of the format {str('[0-9a-fA-F]'): dec(0x_key)}
    """
    h_dict = {}
    ind = 0
    for i in    [str(i) for i in range(10)] + \
                [chr(0x61 + i) for i in range(6)] + \
                [chr(0x41 + i) for i in range(6)]:
        h_dict[i] = ind if ind <= 15 else ind - 6
        ind += 1

    return h_dict


def int_from_string(hex_string):
    """
        Converts hex string to integer.
    """
    hd = hex_dict()

    value = 0
    i = 0
    for char in hex_string[::-1]:
        val = None
        try:
            val = hd[char]
        except KeyError as ke:
            raise HexadecimalStringFormatError("Not a hex string.", "String contained: {0}".format(char))
        value += pow(16, i) * val
        i += 1

    return value


def ip_from_hex(string):
    """
        Converts hex ip string ("1122AA44") to decimal notation (x.x.x.x).
    """
    h_dict = hex_dict()

    rval = ""
    if len(string) != 8:
        raise HexadecimalIpFormatError('Not a valid IPv4 string.', string)

    s = string[::-1]
    o = []
    while s:
        o.append(s[:2])
        s = s[2:]
    for pair in o:
        value = 0
        i = 0
        for h_char in pair:
            try:
                value += pow(16, i) * h_dict[h_char]
            except KeyError:
                raise HexadecimalIpFormatError('Not a valid IPv4 string.', h_char)
            i += 1
        rval += str(value) + '.'
    return rval.rstrip('.')

def port_from_hex(string):
    """
        Converts hex port string to decimal port (0, 65535)
    """

    if len(string) != 4:
        raise HexadecimalPortFormatError('Not a valid hexadecimal port string.', string)

    return str(int_from_string(string))
